import pytest

from .cli import prep_project


def pytest_collect_file(parent, file_path):
    if file_path.name == "pyproject.toml":
        return PyprojectFile.from_parent(parent, path=file_path)


class PyprojectFile(pytest.File):
    def collect(self):
        yield PyprojectItem.from_parent(self, name="")


class PyprojectItem(pytest.Item):
    def runtest(self):
        project = prep_project()
        if project.is_dirty:
            raise ValueError("Project needs to be reconfigured")

    def repr_failure(self, excinfo):
        if isinstance(excinfo.value, ValueError):
            return "\n".join(
                [
                    "Project needs to be reconfigured",
                    "Run 'opinions apply' to fix",
                ]
            )



import json
from pathlib import Path
from typing import Any, Dict, Optional

import tomlkit
from tomlkit.container import Container
from tomlkit.items import Item


class TOMLFile:
    def __init__(self, path: Path):
        self.path = Path(path)
        self._content: Optional[tomlkit.TOMLDocument] = None
        self._dirty = False
        self._initial_content: Optional[str] = None

    @property
    def is_dirty(self) -> bool:
        return self._dirty or (
            self._content is not None
            and self._initial_content is not None
            and tomlkit.dumps(self._content) != self._initial_content
        )

    @property
    def content(self) -> tomlkit.TOMLDocument:
        if self._content is None:
            if self.path.exists():
                with open(self.path) as f:
                    self._content = tomlkit.load(f)
                    self._initial_content = tomlkit.dumps(self._content)
            else:
                self._content = tomlkit.document()
                self._initial_content = ""
        return self._content

    def finalize(self) -> bool:
        wrote_out = False
        if self.is_dirty:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            with open(self.path, "w") as f:
                tomlkit.dump(self.content, f)
            wrote_out = True
        return wrote_out

    def mark_dirty(self):
        self._dirty = True

    def get_or_create_table(self, *path: str, is_super=False) -> tomlkit.TOMLDocument | Item | Container | Any:
        table: Any = self.content
        for path_fragment in path:
            if path_fragment not in table:
                table[path_fragment] = tomlkit.table(is_super_table=is_super)
            table = table[path_fragment]
        return table


class PyProjectTOMLFile(TOMLFile):
    def add_pytest_opt(self, option):
        pytest_iniopts = self.get_or_create_table("tool", "pytest", "ini_options", is_super=True)
        if "addopts" not in pytest_iniopts:
            pytest_iniopts["addopts"] = ""

        if option not in pytest_iniopts["addopts"].split(" "):
            pytest_iniopts["addopts"] += " " + option
            pytest_iniopts["addopts"] = pytest_iniopts["addopts"].strip()


class JsonFile:
    def __init__(self, path: Path):
        self.path = Path(path)
        self._content: Optional[Any] = None
        self._dirty = False
        self._initial_content: Optional[str] = None

    @property
    def is_dirty(self) -> bool:
        return self._dirty or (
            self._content is not None
            and self._initial_content is not None
            and json.dumps(self._content) != self._initial_content
        )

    @property
    def content(self) -> Any:
        if self._content is None:
            if self.path.exists():
                with open(self.path) as f:
                    self._content = json.load(f)
                    self._initial_content = json.dumps(self._content)
            else:
                self._content = {}
                self._initial_content = ""
        return self._content

    def finalize(self) -> bool:
        wrote_out = False
        if self.is_dirty:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            with open(self.path, "w") as f:
                json.dump(self.content, f, indent=4)
            wrote_out = True
        return wrote_out

    def mark_dirty(self):
        self._dirty = True


class OpinionatedProject:
    def __init__(self, base_path: Path):
        self.base_path = base_path
        self._toml_files: Dict[str, TOMLFile] = {}
        self._json_files: Dict[str, JsonFile] = {}

        from opinions.opinion_mypy import MypyOpinion
        from opinions.opinion_pytest_self_check import PytestSelfCheckOpionion
        from opinions.opinion_ruff import RuffOpinion
        from opinions.opinion_ruff_vscode import RuffVSCodeOpinion

        self.opinions = [
            RuffOpinion(self),
            MypyOpinion(self),
            RuffVSCodeOpinion(self),
            PytestSelfCheckOpionion(self),
        ]

    def get_toml_file(self, path: str) -> TOMLFile:
        if path not in self._toml_files:
            self._toml_files[path] = TOMLFile(self.base_path / path)
        return self._toml_files[path]

    def get_json_file(self, path: str) -> JsonFile:
        if path not in self._json_files:
            self._json_files[path] = JsonFile(self.base_path / path)
        return self._json_files[path]

    @property
    def pyproject(self) -> PyProjectTOMLFile:
        if "pyproject.toml" not in self._toml_files:
            self._toml_files["pyproject.toml"] = PyProjectTOMLFile(self.base_path / "pyproject.toml")
        return self._toml_files["pyproject.toml"]  # type: ignore

    def can_manage_project(self) -> bool:
        return (self.base_path / "pyproject.toml").exists()

    def finalize(self) -> bool:
        dirty = False
        for _, toml_file in self._toml_files.items():
            f_dirty = toml_file.finalize()
            dirty = dirty or f_dirty
        for _, json_file in self._json_files.items():
            f_dirty = json_file.finalize()
            dirty = dirty or f_dirty
        return dirty

    @property
    def is_dirty(self):
        return any([x.is_dirty for x in self._toml_files.values()]) or any(
            [x.is_dirty for x in self._json_files.values()]
        )


class Opinion:
    def __init__(self, project: OpinionatedProject):
        self.project = project

    def apply_changes(self):
        raise NotImplementedError()



import argparse
import os
from pathlib import Path

from opinions import OpinionatedProject


def main():
    arg_parser = argparse.ArgumentParser()
    subparsers = arg_parser.add_subparsers(required=True)

    check_parser = subparsers.add_parser("check")
    check_parser.set_defaults(func=check)

    apply_parser = subparsers.add_parser("apply")
    apply_parser.set_defaults(func=apply)

    args = arg_parser.parse_args()
    args.func(args)


def check(args):
    project = prep_project()
    if project.is_dirty:
        print("Changes needed - run 'opinions apply' to fix")
        exit(1)
    else:
        print("No changes needed")


def apply(args):
    project = prep_project()
    result = project.finalize()
    if result:
        print("Project configured!")
    else:
        print("Nothing to do")


def prep_project() -> OpinionatedProject:
    project = OpinionatedProject(Path(os.getcwd()))
    for opinion in project.opinions:
        opinion.apply_changes()

    return project



from . import Opinion


class RuffOpinion(Opinion):
    def apply_changes(self):
        ruff_tool = self.project.pyproject.get_or_create_table("tool", "ruff", is_super=True)
        ruff_tool["line-length"] = 120

        if "select" not in ruff_tool:
            ruff_tool["select"] = []

        for lint_rule in [
            "E4",
            "E7",
            "E9",
            "F",
            "UP",
            "S",
            "A",
            "I",
        ]:
            if lint_rule not in ruff_tool["select"]:
                ruff_tool["select"].append(lint_rule)

        self.project.pyproject.add_pytest_opt("--ruff")
        self.project.pyproject.add_pytest_opt("--ruff-format")



from typing import List

from . import Opinion


class RuffVSCodeOpinion(Opinion):
    def apply_changes(self) -> None:
        vscode_settings = self.project.get_json_file(".vscode/settings.json")

        vscode_settings.content["mypy-type-checker.importStrategy"] = "fromEnvironment"

        if "[python]" not in vscode_settings.content:
            vscode_settings.content["[python]"] = {}
        python_settings = vscode_settings.content["[python]"]

        python_settings["editor.defaultFormatter"] = "charliermarsh.ruff"
        python_settings["editor.formatOnSave"] = True

        if "editor.codeActionsOnSave" not in python_settings:
            python_settings["editor.codeActionsOnSave"] = {}
        on_save_settings = python_settings["editor.codeActionsOnSave"]
        on_save_settings["source.organizeImports.ruff"] = "explicit"

        extensions_config = self.project.get_json_file(".vscode/extensions.json")
        if "recommendations" not in extensions_config.content:
            extensions_config.content["recommendations"] = []
        recommendations: List[str] = extensions_config.content["recommendations"]

        for ext in [
            "ms-python.python",
            "charliermarsh.ruff",
            "ms-python.mypy-type-checker",
        ]:
            if ext not in recommendations:
                recommendations.append(ext)



from . import Opinion


class PytestSelfCheckOpionion(Opinion):
    def apply_changes(self):
        if self.project.pyproject.content.get("tool", {}).get("poetry", {}).get("name", None) != "opinions":
            self.project.pyproject.add_pytest_opt("--opinions")
